@@ -1,13 +1,15 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
-#include "prelude.h"
 #include "bfstd.h"
+
+#include "bfs.h"
 #include "bit.h"
 #include "diag.h"
 #include "sanity.h"
 #include "thread.h"
 #include "xregex.h"
+
 #include <errno.h>
 #include <fcntl.h>
 #include <langinfo.h>
@@ -27,13 +29,13 @@
 #include <unistd.h>
 #include <wchar.h>
 
-#if BFS_USE_SYS_SYSMACROS_H
+#if __has_include(<sys/sysmacros.h>)
 #  include <sys/sysmacros.h>
-#elif BFS_USE_SYS_MKDEV_H
+#elif __has_include(<sys/mkdev.h>)
 #  include <sys/mkdev.h>
 #endif
 
-#if BFS_USE_UTIL_H
+#if __has_include(<util.h>)
 #  include <util.h>
 #endif
 
@@ -324,7 +326,7 @@ const char *xstrerror(int errnum) {
 
 	// On FreeBSD with MemorySanitizer, duplocale() triggers
 	// https://github.com/llvm/llvm-project/issues/65532
-#if BFS_HAS_STRERROR_L && !(__FreeBSD__ && SANITIZE_MEMORY)
+#if BFS_HAS_STRERROR_L && !(__FreeBSD__ && __SANITIZE_MEMORY__)
 #  if BFS_HAS_USELOCALE
 	locale_t loc = uselocale((locale_t)0);
 #  else
@@ -715,14 +717,14 @@ int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *
 }
 
 long xsysconf(int name) {
-#if __FreeBSD__ && SANITIZE_MEMORY
+#if __FreeBSD__ && __SANITIZE_MEMORY__
 	// Work around https://github.com/llvm/llvm-project/issues/88163
 	__msan_scoped_disable_interceptor_checks();
 #endif
 
 	long ret = sysconf(name);
 
-#if __FreeBSD__ && SANITIZE_MEMORY
+#if __FreeBSD__ && __SANITIZE_MEMORY__
 	__msan_scoped_enable_interceptor_checks();
 #endif
 
