@@ -1,13 +1,12 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
-#include "prelude.h"
 #include "tests.h"
-#include "ioq.h"
-#include "bfstd.h"
+
 #include "diag.h"
 #include "dir.h"
-#include <errno.h>
+#include "ioq.h"
+
 #include <fcntl.h>
 #include <stdlib.h>
 
